/*
 * Print audit records
 *
 * Copyright (C) 2003, SuSE Linux AG
 * Written by okir@suse.de
 */

#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <pwd.h>

#include <linux/unistd.h>
#include <linux/sys.h>

#include <laus.h>
#include <laussrv.h>
#include "syscall.h"

static void	print_syscall_msg(struct aud_message *msg);
static void	print_login_msg(struct aud_message *msg);
static void	print_text_msg(struct aud_message *msg);
static void	print_netlink_msg(struct aud_message *msg);
static void	print_exit_msg(struct aud_message *msg);

int
audit_print(struct aud_message *msg, int flags)
{
	audit_print_header(msg, flags);
	switch (msg->msg_type) {
	case AUDIT_MSG_LOGIN:
		print_login_msg(msg); break;
	case AUDIT_MSG_TEXT:
		print_text_msg(msg); break;
	case AUDIT_MSG_SYSCALL:
		print_syscall_msg(msg); break;
	case AUDIT_MSG_NETLINK:
		print_netlink_msg(msg); break;
	case AUDIT_MSG_EXIT:
		print_exit_msg(msg); break;
	default:
		printf("[unknown message type %u]", msg->msg_type);
	}

	printf("\n");
	return 0;
}

void
audit_print_caption(int flags)
{
	switch (flags & AUDPR_TIME) {
	case AUDPR_TIME_FMT_ISO8601:
		printf("     TIME            "); break;
	case AUDPR_TIME_FMT_UNIX:
		printf("     TIME           "); break;
	case AUDPR_TIME_FMT_RAW:
		printf("   TIME      "); break;
	}
	if (flags & AUDPR_SEQNR)
		printf("SEQNR   ");
	if (flags & AUDPR_PROCESSID)
		printf("PID     ");
	if (flags & AUDPR_LOGINID)
		printf("LOGIN   ");
	if (flags & AUDPR_UID_ALL){
		printf("EUID   ");
		printf("SUID   ");
		printf("RUID   ");
		printf("FSUID  ");
	}
	if (flags & AUDPR_GID_ALL) {
		printf("EGID   ");
		printf("RGID   ");
		printf("SGID   ");
		printf("FSGID   ");
	}
	if (flags & AUDPR_AUDITID)
		printf("SESS   ");
	printf("DATA\n");
}

void
audit_print_header(struct aud_message *msg, int flags)
{
	if (flags & AUDPR_TIME) {
		time_t		timestamp = msg->msg_timestamp;
		struct tm	*tm;
		char		timestr[64];

		tm = localtime(&timestamp);

		switch (flags & AUDPR_TIME) {
		case AUDPR_TIME_FMT_ISO8601:
			strftime(timestr, sizeof(timestr), "%Y-%m-%dT%H:%M:%S", tm);
			break;
		case AUDPR_TIME_FMT_UNIX:
			strftime(timestr, sizeof(timestr), "%d %b %y %H:%M:%S", tm);
			break;
		case AUDPR_TIME_FMT_RAW:
			snprintf(timestr, sizeof(timestr), "%-8lu ",
					(unsigned long) timestamp);
			break;
		default:
			strcpy(timestr, "-");
		}
		printf("%s ", timestr);
	}

	if (flags & AUDPR_SEQNR)
		printf("%6d ",msg->msg_seqnr);
	if (flags & AUDPR_PROCESSID)
		printf("%6d ", msg->msg_pid);
	if (flags & AUDPR_LOGINID) {
		struct passwd	*pw;

		if ((pw = getpwuid(msg->msg_login_uid)) == NULL)
			printf("%8d ", msg->msg_login_uid);
		else
			printf("%8s ", pw->pw_name);
	}

	if (flags & AUDPR_UID_ALL){
		printf("%6u ", msg->msg_euid);
		printf("%6u ", msg->msg_suid);
		printf("%6u ", msg->msg_ruid);
		printf("%6u ", msg->msg_fsuid);
	}
	if (flags & AUDPR_GID_ALL){
		printf("%6u ", msg->msg_egid);
		printf("%6u ", msg->msg_rgid);
		printf("%6u ", msg->msg_sgid);
		printf("%6u  ", msg->msg_fsgid);
	}
	if (flags & AUDPR_AUDITID)
		printf("%5u    ", msg->msg_audit_id);

	/* XXX handle other flags */

	if ((flags & AUDPR_EVNAME) && msg->msg_evname[0])
		printf("[%.*s] ", AUD_MAX_EVNAME, msg->msg_evname);
}
	

void
print_login_msg(struct aud_message *msg)
{
	struct aud_msg_login	*logmsg;

	if (msg->msg_size != sizeof(*msg) + sizeof(*logmsg)) {
		printf("login message, strange message size\n");
		return;
	}

	logmsg = (struct aud_msg_login *) &msg->msg_data;
	printf("LOGIN: uid=%u", logmsg->uid);
	if (logmsg->hostname[0])
		printf(", hostname=%s", logmsg->hostname);
	if (logmsg->address[0])
		printf(", address=%s", logmsg->address);
	if (logmsg->terminal[0])
		printf(", terminal=%s", logmsg->terminal);
	if (logmsg->executable[0])
		printf(", executable=%s", logmsg->executable);
}

void
print_text_msg(struct aud_message *msg)
{
	int	len;

	len = msg->msg_size - sizeof(*msg);
	if (len <= 0) {
		printf("(empty text message)");
		return;
	}

	printf("%.*s", len, (char *) &msg->msg_data);
}

void
print_netlink_msg(struct aud_message *msg)
{
	struct aud_msg_netlink *nlmsg;
	int		len;

	len = msg->msg_size - sizeof(*msg);
	if (len <= sizeof(*nlmsg)) {
		printf("(empty netlink message)");
		return;
	}

	nlmsg = (struct aud_msg_netlink *) &msg->msg_data;
	len -= sizeof(*nlmsg);

	printf("NETLINK");
	rtnetlink_print(nlmsg, len);
}

/*
 * System call audit message
 */
void
print_syscall_msg(struct aud_message *msg)
{
	struct syscall_data	data;
	struct aud_msg_syscall	*aud;
	caddr_t			args;
	unsigned int		size;
	u_int32_t		type, len;
	unsigned int		nargs;

	aud  = (struct aud_msg_syscall *) &msg->msg_data;
	args = aud->data;
	size = aud->length;

	memset(&data, 0, sizeof(data));
	data.major  = aud->major;
	data.minor  = aud->minor;
	data.result = aud->result;

	for (nargs = 0; size >= 8 && nargs < MAX_ARGS; nargs++) {
		struct syscall_arg	*arg = &data.args[nargs];

		memcpy(&type, args, 4); args += 4;
		memcpy(&len,  args, 4); args += 4;
		size -= 8;

		if (type == AUDIT_ARG_END)
			break;
		/* Do not display the NUL byte */
		if (type == AUDIT_ARG_STRING && len && args[len-1] == '\0')
			len--;

		arg->type = type;
		arg->data = args;
		arg->len  = len;

		args += len;
		size -= len;
	}
	data.nargs = nargs;

	syscall_print(&data);
}

void
print_exit_msg(struct aud_message *msg) 
{
	struct aud_msg_exit *aud;
	aud  = (struct aud_msg_exit *) &msg->msg_data;
	printf ("EXIT: %ld", aud->code);
}
